@@ -30,7 +30,6 @@
 #include <gprs_debug.h>
 #include <cxx_linuxlist.h>
 #include <pdch.h>
-#include <gprs_ms_storage.h>
 #include <sba.h>
 #include <bts_pch_timer.h>
 
@@ -53,7 +52,6 @@ extern "C" {
 #include <errno.h>
 #include <string.h>
 
-#define RFN_MODULUS 42432
 #define RFN_THRESHOLD RFN_MODULUS / 2
 
 extern void *tall_pcu_ctx;
@@ -79,9 +77,11 @@ void bts_trx_free_all_tbf(struct gprs_rlcmac_trx *trx)
 
 static struct osmo_tdef T_defs_bts[] = {
 	{ .T=3142, .default_val=20,  .unit=OSMO_TDEF_S,  .desc="Wait Indication used in Imm Ass Reject during TBF Establishment (CCCH)", .val=0, .min_val = 0, .max_val = 255 }, /* TS 44.018 10.5.2.43, TS 44.060 7.1.3.2.1 (T3172) */
+	{ .T=3168, .default_val=4000, .unit=OSMO_TDEF_MS,  .desc="Time MS waits for PACKET UPLINK ACK when establishing a UL TBF", .val=0 },
 	{ .T=3169, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of USF and TFI(s) after the MS uplink TBF assignment is invalid", .val=0 },
 	{ .T=3191, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of TFI(s) after sending (1) last RLC Data Block on TBF(s), or (2) PACKET TBF RELEASE for an MBMS radio bearer", .val=0 },
-	{ .T=3193, .default_val=100, .unit=OSMO_TDEF_MS, .desc="Reuse of TFI(s) after reception of final PACKET DOWNLINK ACK/NACK from MS for TBF", .val=0 },
+	{ .T=3192, .default_val=1500, .unit=OSMO_TDEF_MS, .desc="Time MS stays monitoring the PDCH after transmitting DL ACK/NACK for last DL block (FBI=1). Configured at the BSC (SI13).", .val=0 },
+	{ .T=3193, .default_val=1600, .unit=OSMO_TDEF_MS, .desc="Reuse of TFI(s) after reception of final PACKET DOWNLINK ACK/NACK from MS for TBF", .val=0 },
 	{ .T=3195, .default_val=5,   .unit=OSMO_TDEF_S,  .desc="Reuse of TFI(s) upon no response from the MS (radio failure or cell change) for TBF/MBMS radio bearer", .val=0 },
 	{ .T = -16, .default_val = 1000, .unit = OSMO_TDEF_MS,
 		.desc = "Granularity for *:all_allocated rate counters: amount of milliseconds that one counter increment"
@@ -241,10 +241,9 @@ static const struct osmo_stat_item_group_desc bts_statg_desc = {
 
 static int bts_talloc_destructor(struct gprs_rlcmac_bts* bts)
 {
-	/* this can cause counter updates and must not be left to the
-	 * m_ms_store's destructor */
-	bts->ms_store->cleanup();
-	delete bts->ms_store;
+	struct GprsMs *ms;
+	while ((ms = llist_first_entry_or_null(&bts->ms_list, struct GprsMs, list)))
+		talloc_free(ms);
 
 	gprs_bssgp_destroy(bts);
 
@@ -282,8 +281,6 @@ struct gprs_rlcmac_bts* bts_alloc(struct gprs_pcu *pcu, uint8_t bts_nr)
 	bts->pcu = pcu;
 	bts->nr = bts_nr;
 
-	bts->ms_store = new GprsMsStorage(bts);
-
 	bts->cur_fn = FN_UNSET;
 	bts->cur_blk_fn = -1;
 	bts->max_cs_dl = MAX_GPRS_CS;
@@ -332,6 +329,7 @@ struct gprs_rlcmac_bts* bts_alloc(struct gprs_pcu *pcu, uint8_t bts_nr)
 	llist_add_tail(&bts->list, &pcu->bts_list);
 
 	INIT_LLIST_HEAD(&bts->pch_timer);
+	INIT_LLIST_HEAD(&bts->ms_list);
 
 	return bts;
 }
@@ -460,7 +458,7 @@ int bts_add_paging(struct gprs_rlcmac_bts *bts, const struct paging_req_cs *req,
 	 * Don't mark, if TBF uses a different slot that is already marked. */
 	memset(slot_mask, 0, sizeof(slot_mask));
 
-	llist_for_each(tmp, bts_ms_store(bts)->ms_list()) {
+	llist_for_each(tmp, &bts->ms_list) {
 		ms = llist_entry(tmp, typeof(*ms), list);
 		struct gprs_rlcmac_tbf *tbfs[] = { ms->ul_tbf, ms->dl_tbf };
 		for (l = 0; l < ARRAY_SIZE(tbfs); l++) {
@@ -535,7 +533,7 @@ void bts_send_gsmtap_rach(struct gprs_rlcmac_bts *bts,
 	}
 
 	bts_send_gsmtap_meas(bts, categ, true, rip->trx_nr, rip->ts_nr, channel,
-			 bts_rfn_to_fn(bts, rip->rfn), ra_buf,
+			 rip->fn, ra_buf,
 			 rip->is_11bit ? 2 : 1, &meas);
 }
 
@@ -668,13 +666,10 @@ int bts_tfi_find_free(const struct gprs_rlcmac_bts *bts, enum gprs_rlcmac_tbf_di
 	return best_first_tfi;
 }
 
-int bts_rcv_imm_ass_cnf(struct gprs_rlcmac_bts *bts, const uint8_t *data, uint32_t fn)
+static int tlli_from_imm_ass(uint32_t *tlli, const uint8_t *data)
 {
-	struct gprs_rlcmac_dl_tbf *dl_tbf;
 	const struct gsm48_imm_ass *imm_ass = (struct gsm48_imm_ass *)data;
 	uint8_t plen;
-	uint32_t tlli;
-	GprsMs *ms;
 
 	/* Move to IA Rest Octets: TS 44.018 9.1.18 "The L2 pseudo length of
 	 * this message is the sum of lengths of all information elements
@@ -685,31 +680,60 @@ int bts_rcv_imm_ass_cnf(struct gprs_rlcmac_bts *bts, const uint8_t *data, uint32
 	data += 1 + plen;
 
 	if ((*data & 0xf0) != 0xd0) {
-		LOGP(DTBFDL, LOGL_ERROR, "FN=%u Got IMM.ASS confirm, but rest "
+		LOGP(DTBFDL, LOGL_ERROR, "Got IMM.ASS confirm, but rest "
 			"octets do not start with bit sequence 'HH01' "
-			"(Packet Downlink Assignment)\n", fn);
+			"(Packet Downlink Assignment)\n");
 		return -EINVAL;
 	}
 
 	/* get TLLI from downlink assignment */
-	tlli = (uint32_t)((*data++) & 0xf) << 28;
-	tlli |= (*data++) << 20;
-	tlli |= (*data++) << 12;
-	tlli |= (*data++) << 4;
-	tlli |= (*data++) >> 4;
+	*tlli = (uint32_t)((*data++) & 0xf) << 28;
+	*tlli |= (*data++) << 20;
+	*tlli |= (*data++) << 12;
+	*tlli |= (*data++) << 4;
+	*tlli |= (*data++) >> 4;
 
-	ms = bts_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
+	return 0;
+}
+
+int bts_rcv_imm_ass_cnf(struct gprs_rlcmac_bts *bts, const uint8_t *data, uint32_t tlli)
+{
+	struct gprs_rlcmac_dl_tbf *dl_tbf;
+	GprsMs *ms;
+	int rc;
+
+	/* NOTE: A confirmation for a downlink IMMEDIATE ASSIGNMENT can be received using two different methods.
+	 * One way is to send the whole IMMEDIATE ASSIGNMENT back to the PCU and the TLLI, which we use as
+	 * reference is extracted from the rest octets of this message. Alternatively the TLLI may be sent as
+	 * confirmation directly. */
+
+	/* Extract TLLI from the presented IMMEDIATE ASSIGNMENT
+	 * (if present and only when TLLI that is supplied as function parameter is valid.) */
+	if (data && tlli == GSM_RESERVED_TMSI) {
+		rc = tlli_from_imm_ass(&tlli, data);
+		if (rc != 0)
+			return -EINVAL;
+	}
+
+	/* Make sure TLLI is valid */
+	if (tlli == GSM_RESERVED_TMSI) {
+		LOGP(DTBFDL, LOGL_ERROR, "Got IMM.ASS confirm, but TLLI is invalid!\n");
+		return -EINVAL;
+	}
+
+	/* Find related TBF and send confirmation signal to FSM */
+	ms = bts_get_ms_by_tlli(bts, tlli, GSM_RESERVED_TMSI);
 	if (!ms) {
-		LOGP(DTBFDL, LOGL_ERROR, "FN=%u Got IMM.ASS confirm for unknown MS with TLLI=%08x\n", fn, tlli);
+		LOGP(DTBFDL, LOGL_ERROR, "Got IMM.ASS confirm for unknown MS with TLLI=%08x\n", tlli);
 		return -EINVAL;
 	}
 	dl_tbf = ms_dl_tbf(ms);
 	if (!dl_tbf) {
-		LOGPMS(ms, DTBFDL, LOGL_ERROR, "FN=%u Got IMM.ASS confirm, but MS has no DL TBF!\n", fn);
+		LOGPMS(ms, DTBFDL, LOGL_ERROR, "Got IMM.ASS confirm, but MS has no DL TBF!\n");
 		return -EINVAL;
 	}
 
-	LOGPTBFDL(dl_tbf, LOGL_DEBUG, "FN=%u Got IMM.ASS confirm\n", fn);
+	LOGPTBFDL(dl_tbf, LOGL_DEBUG, "Got IMM.ASS confirm\n");
 	osmo_fsm_inst_dispatch(dl_tbf->state_fi, TBF_EV_ASSIGN_PCUIF_CNF, NULL);
 
 	return 0;
@@ -718,32 +742,27 @@ int bts_rcv_imm_ass_cnf(struct gprs_rlcmac_bts *bts, const uint8_t *data, uint32
 /* Determine the full frame number from a relative frame number */
 uint32_t bts_rfn_to_fn(const struct gprs_rlcmac_bts *bts, uint32_t rfn)
 {
-	uint32_t m_cur_rfn;
+	uint32_t m_cur_fn, m_cur_rfn;
 	uint32_t fn_rounded;
 
-	/* make sure RFN does not exceed the maximum possible value of a valid
-	 * GSM frame number. */
-	OSMO_ASSERT(rfn < GSM_MAX_FN);
-
-	/* Note: If a BTS is sending in a rach request it will be fully aware
-	 * of the frame number. If the PCU is used in a BSC-co-located setup.
-	 * The BSC will forward the incoming RACH request. The RACH request
-	 * only contains the relative frame number (Fn % 42432) in its request
-	 * reference. This PCU implementation has to fit both scenarios, so
-	 * we need to assume that Fn is a relative frame number. */
-
 	/* Ensure that all following calculations are performed with the
 	 * relative frame number */
-	if (rfn >= RFN_MODULUS)
+	OSMO_ASSERT(rfn < RFN_MODULUS);
+
+	m_cur_fn = bts_current_frame_number(bts);
+	if (OSMO_UNLIKELY(m_cur_fn == FN_UNSET)) {
+		LOGP(DRLCMAC, LOGL_ERROR, "Unable to calculate full FN from RFN %u: Current FN not known!\n",
+		     rfn);
 		return rfn;
+	}
 
 	/* Compute an internal relative frame number from the full internal
 	   frame number */
-	m_cur_rfn = bts->cur_fn % RFN_MODULUS;
+	m_cur_rfn = fn2rfn(m_cur_fn);
 
 	/* Compute a "rounded" version of the internal frame number, which
 	 * exactly fits in the RFN_MODULUS raster */
-	fn_rounded = GSM_TDMA_FN_SUB(bts->cur_fn, m_cur_rfn);
+	fn_rounded = GSM_TDMA_FN_SUB(m_cur_fn, m_cur_rfn);
 
 	/* If the delta between the internal and the external relative frame
 	 * number exceeds a certain limit, we need to assume that the incoming
@@ -752,7 +771,7 @@ uint32_t bts_rfn_to_fn(const struct gprs_rlcmac_bts *bts, uint32_t rfn)
 	if (GSM_TDMA_FN_DIFF(rfn, m_cur_rfn) > RFN_THRESHOLD) {
 		LOGP(DRLCMAC, LOGL_DEBUG,
 		     "Race condition between rfn (%u) and m_cur_fn (%u) detected: rfn belongs to the previous modulus %u cycle, wrapping...\n",
-		     rfn, bts->cur_fn, RFN_MODULUS);
+		     rfn, m_cur_fn, RFN_MODULUS);
 		if (fn_rounded < RFN_MODULUS) {
 			LOGP(DRLCMAC, LOGL_DEBUG,
 			"Cornercase detected: wrapping crosses %u border\n",
@@ -926,15 +945,13 @@ int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 	if (rip->is_11bit)
 		bts_do_rate_ctr_inc(bts, CTR_RACH_REQUESTS_11BIT);
 
-	/* Determine full frame number */
-	uint32_t Fn = bts_rfn_to_fn(bts, rip->rfn);
 	uint8_t ta = qta2ta(rip->qta);
 
 	bts_send_gsmtap_rach(bts, PCU_GSMTAP_C_UL_RACH, GSMTAP_CHANNEL_RACH, rip);
 
 	LOGP(DRLCMAC, LOGL_DEBUG, "MS requests Uplink resource on CCCH/RACH: "
 	     "ra=0x%02x (%d bit) Fn=%u qta=%d\n", rip->ra,
-	     rip->is_11bit ? 11 : 8, Fn, rip->qta);
+	     rip->is_11bit ? 11 : 8, rip->fn, rip->qta);
 
 	/* Parse [EGPRS Packet] Channel Request from RACH.ind */
 	rc = parse_rach_ind(rip, &chan_req);
@@ -953,6 +970,12 @@ int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 		LOGP(DRLCMAC, LOGL_DEBUG, "MS requests single TS uplink transmission "
 		     "(one phase packet access)\n");
 		if (bts->pcu->vty.force_two_phase) {
+			/* 3GPP TS 44.018 3.5.2.1.3.1: "If the establishment cause in the
+			 * CHANNEL REQUEST message indicates a request for one phase packet access,
+			 * the network may grant either a one phase packet access or a single block
+			 * packet access for the mobile station. If a single block packet access is
+			 * granted, it forces the mobile station to perform a two phase packet access."
+			 */
 			LOGP(DRLCMAC, LOGL_DEBUG, "Forcing two phase access\n");
 			chan_req.single_block = true;
 		}
@@ -980,8 +1003,13 @@ int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 		     "SBFn=%u TRX=%u TS=%u\n", sb_fn, pdch->trx->trx_no, pdch->ts_no);
 		bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_UL_TBF_TWO_PHASE);
 	} else {
-		GprsMs *ms = bts_alloc_ms(bts, 0, chan_req.egprs_mslot_class);
+		GprsMs *ms = ms_alloc(bts, __func__);
+		ms_set_egprs_ms_class(ms, chan_req.egprs_mslot_class);
 		tbf = ms_new_ul_tbf_assigned_agch(ms);
+		/* Here either tbf was created and it holds a ref to MS, or tbf
+		 * creation failed and MS will end up without references and being
+		 * freed: */
+		ms_unref(ms, __func__);
 		if (!tbf) {
 			/* Send RR Immediate Assignment Reject */
 			rc = -EBUSY;
@@ -1000,12 +1028,15 @@ int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 	     trx->trx_no, trx->arfcn & ~ARFCN_FLAG_MASK,
 	     pdch->ts_no, ta, pdch->tsc, tbf ? tbf->tfi() : -1, usf);
 	plen = Encoding::write_immediate_assignment(pdch, tbf, bv,
-		false, rip->ra, Fn, ta, usf, false, sb_fn,
+		false, rip->ra, rip->rfn, ta, usf, false, fn2rfn(sb_fn),
 		bts_get_ms_pwr_alpha(bts), bts->pcu->vty.gamma, -1,
 		rip->burst_type);
 	bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_UL_TBF);
 	if (plen >= 0) {
-		pcu_l1if_tx_agch(bts, bv, plen);
+		if (the_pcu->pcu_if_version >= 0x0b)
+			pcu_l1if_tx_agch2(bts, bv, plen, false, GSM_RESERVED_TMSI);
+		else
+			pcu_l1if_tx_agch(bts, bv, plen);
 		rc = 0;
 	} else {
 		rc = plen;
@@ -1016,11 +1047,15 @@ int bts_rcv_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 send_imm_ass_rej:
 	LOGP(DRLCMAC, LOGL_DEBUG, "Tx Immediate Assignment Reject on AGCH\n");
 	plen = Encoding::write_immediate_assignment_reject(
-		bv, rip->ra, Fn, rip->burst_type,
+		bv, rip->ra, rip->rfn, rip->burst_type,
 		(uint8_t)osmo_tdef_get(bts->T_defs_bts, 3142, OSMO_TDEF_S, -1));
 	bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_REJ);
-	if (plen >= 0)
-		pcu_l1if_tx_agch(bts, bv, plen);
+	if (plen >= 0) {
+		if (the_pcu->pcu_if_version >= 0x0b)
+			pcu_l1if_tx_agch2(bts, bv, plen, false, GSM_RESERVED_TMSI);
+		else
+			pcu_l1if_tx_agch(bts, bv, plen);
+	}
 	bitvec_free(bv);
 	/* rc was already properly set before goto */
 	return rc;
@@ -1036,7 +1071,7 @@ static uint32_t ptcch_slot_map[PTCCH_TAI_NUM] = {
 
 int bts_rcv_ptcch_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params *rip)
 {
-	uint32_t fn416 = bts_rfn_to_fn(bts, rip->rfn) % 416;
+	uint16_t fn416 = rip->rfn % 416;
 	struct gprs_rlcmac_pdch *pdch;
 	uint8_t ss;
 
@@ -1077,14 +1112,13 @@ int bts_rcv_ptcch_rach(struct gprs_rlcmac_bts *bts, const struct rach_ind_params
 
 void bts_snd_dl_ass(struct gprs_rlcmac_bts *bts, const struct gprs_rlcmac_dl_tbf *tbf)
 {
-	uint16_t pgroup = ms_paging_group(tbf_ms(tbf));
 	int plen;
-	const struct gprs_rlcmac_pdch *pdch;
-
-	LOGPTBF(tbf, LOGL_INFO, "TX: START Immediate Assignment Downlink (PCH)\n");
 
 	/* Only one TS can be assigned through PCH, hence the first one is the only one: */
-	pdch = tbf_get_first_ts_const(tbf);
+	const struct gprs_rlcmac_pdch *pdch = tbf_get_first_ts_const(tbf);
+	OSMO_ASSERT(pdch);
+
+	LOGPTBFDL(tbf, LOGL_INFO, "Tx CCCH (PCH) Immediate Assignment [PktDlAss=%s] TA=%d\n", pdch_name(pdch), tbf->ta());
 
 	bitvec *immediate_assignment = bitvec_alloc(22, tall_pcu_ctx); /* without plen */
 	bitvec_unhex(immediate_assignment, DUMMY_VEC); /* standard '2B'O padding */
@@ -1093,17 +1127,34 @@ void bts_snd_dl_ass(struct gprs_rlcmac_bts *bts, const struct gprs_rlcmac_dl_tbf
 	 * that the resource reference cannot be confused with any CHANNEL REQUEST
 	 * message sent by a mobile station.  Use last_rts_fn + 21216 (16 TDMA
 	 * super-frame periods, or ~21.3 seconds) to achieve a decent distance. */
-	LOGP(DRLCMAC, LOGL_DEBUG, " - TRX=%d (%d) TS=%d TA=%d\n",
-	     tbf->trx->trx_no, tbf->trx->arfcn, pdch->ts_no, tbf->ta());
 	plen = Encoding::write_immediate_assignment(pdch,
 						    tbf, immediate_assignment, true, 125,
-						    GSM_TDMA_FN_SUM(pdch->last_rts_fn, 21216),
+						    fn2rfn(GSM_TDMA_FN_SUM(pdch->last_rts_fn, 21216)),
 						    tbf->ta(), 7, false, 0,
 						    bts_get_ms_pwr_alpha(bts), bts->pcu->vty.gamma, -1,
 						    GSM_L1_BURST_TYPE_ACCESS_0);
 	if (plen >= 0) {
 		bts_do_rate_ctr_inc(bts, CTR_IMMEDIATE_ASSIGN_DL_TBF);
-		pcu_l1if_tx_pch(bts, immediate_assignment, plen, pgroup);
+
+		if (the_pcu->pcu_if_version >= 0x0b) {
+			if (ms_imsi_is_valid(tbf->ms())) {
+				pcu_l1if_tx_pch2(bts, immediate_assignment, plen, true, tbf->imsi(), tbf->tlli());
+			} else {
+				/* During GMM ATTACH REQUEST, the IMSI is not yet known to the PCU or SGSN. (It is
+				 * requested after the GMM ATTACH REQUEST with the GMM IDENTITY REQUEST.) When the PCU
+				 * has to assign a DL TBF but the IMSI is not known, then the IMMEDIATE ASSIGNMENT is
+				 * sent on the AGCH. The reason for this is that without an IMSI we can not calculate
+				 + the paging group, which would be necessary for transmission on PCH. Since the IMSI
+				 * is usually only unknown during the GMM ATTACH REQUEST, we may assume that the MS
+				 * is in non-DRX mode and hence it is listening on all CCCH blocks, including AGCH.
+				 *
+				 * See also: 3gpp TS 44.060, section 5.5.1.5
+				 *           3gpp TS 45.002, section 6.5.3, 6.5.6 */
+				pcu_l1if_tx_agch2(bts, immediate_assignment, plen, true, tbf->tlli());
+			}
+		} else {
+			pcu_l1if_tx_pch(bts, immediate_assignment, plen, tbf->imsi());
+		}
 	}
 
 	bitvec_free(immediate_assignment);
@@ -1168,26 +1219,42 @@ bool bts_cs_dl_is_supported(const struct gprs_rlcmac_bts* bts, CodingScheme cs)
 	}
 }
 
-GprsMs *bts_alloc_ms(struct gprs_rlcmac_bts* bts, uint8_t ms_class, uint8_t egprs_ms_class)
+struct GprsMs *bts_get_ms(const struct gprs_rlcmac_bts *bts, uint32_t tlli, uint32_t old_tlli,
+			  const char *imsi)
 {
-	GprsMs *ms;
-	ms = bts_ms_store(bts)->create_ms();
+	struct llist_head *tmp;
 
-	ms_set_timeout(ms, osmo_tdef_get(bts->pcu->T_defs, -2030, OSMO_TDEF_S, -1));
-	ms_set_ms_class(ms, ms_class);
-	ms_set_egprs_ms_class(ms, egprs_ms_class);
+	if (tlli != GSM_RESERVED_TMSI || old_tlli != GSM_RESERVED_TMSI) {
+		llist_for_each(tmp, &bts->ms_list) {
+			struct GprsMs *ms = llist_entry(tmp, typeof(*ms), list);
+			if (ms_check_tlli(ms, tlli))
+				return ms;
+			if (ms_check_tlli(ms, old_tlli))
+				return ms;
+		}
+	}
 
-	return ms;
+	/* not found by TLLI */
+
+	if (imsi && imsi[0] != '\0') {
+		llist_for_each(tmp, &bts->ms_list) {
+			struct GprsMs *ms = llist_entry(tmp, typeof(*ms), list);
+			if (ms_imsi_is_valid(ms) && strcmp(imsi, ms_imsi(ms)) == 0)
+				return ms;
+		}
+	}
+
+	return NULL;
 }
 
-struct GprsMsStorage *bts_ms_store(const struct gprs_rlcmac_bts *bts)
+struct GprsMs *bts_get_ms_by_tlli(const struct gprs_rlcmac_bts *bts, uint32_t tlli, uint32_t old_tlli)
 {
-	return bts->ms_store;
+	return bts_get_ms(bts, tlli, old_tlli, NULL);
 }
 
-struct GprsMs *bts_ms_by_tlli(struct gprs_rlcmac_bts *bts, uint32_t tlli, uint32_t old_tlli)
+struct GprsMs *bts_get_ms_by_imsi(const struct gprs_rlcmac_bts *bts, const char *imsi)
 {
-	return bts_ms_store(bts)->get_ms(tlli, old_tlli);
+	return bts_get_ms(bts, GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, imsi);
 }
 
 /* update TA based on TA provided by PH-DATA-IND */
@@ -1395,16 +1462,6 @@ void bts_recalc_max_mcs(struct gprs_rlcmac_bts *bts)
 	bts_set_max_mcs_ul(bts, mcs_ul);
 }
 
-struct GprsMs *bts_ms_by_imsi(struct gprs_rlcmac_bts *bts, const char *imsi)
-{
-	return bts_ms_store(bts)->get_ms(GSM_RESERVED_TMSI, GSM_RESERVED_TMSI, imsi);
-}
-
-const struct llist_head* bts_ms_list(struct gprs_rlcmac_bts *bts)
-{
-	return bts_ms_store(bts)->ms_list();
-}
-
 uint8_t bts_get_ms_pwr_alpha(const struct gprs_rlcmac_bts *bts)
 {
 	if (bts->pcu->vty.force_alpha != (uint8_t)-1)
