@@ -28,9 +28,9 @@
 #include <gprs_debug.h>
 #include <gprs_ms.h>
 #include <pcu_utils.h>
-#include <gprs_ms_storage.h>
 #include <sba.h>
 #include <pdch.h>
+#include <alloc_algo.h>
 
 extern "C" {
 #include <osmocom/core/msgb.h>
@@ -128,6 +128,11 @@ gprs_rlcmac_tbf::gprs_rlcmac_tbf(struct gprs_rlcmac_bts *bts_, GprsMs *ms, gprs_
 	m_name_buf[0] = '\0';
 
 	m_created_ts = time(NULL);
+	/* set timestamp */
+	osmo_clock_gettime(CLOCK_MONOTONIC, &meas.rssi_tv);
+
+	m_ctrs = rate_ctr_group_alloc(this, &tbf_ctrg_desc, next_tbf_ctr_group_id++);
+	OSMO_ASSERT(m_ctrs);
 }
 
 
@@ -217,8 +222,10 @@ void tbf_unlink_pdch(struct gprs_rlcmac_tbf *tbf)
 	 * confirmation from the MS and goes through the FLOW state. Hence, we
 	 * may have ULC pollings ongoing and we need to make sure we drop all
 	 * reserved nodes there: */
-	if (tbf->control_ts)
+	if (tbf->control_ts) {
 		pdch_ulc_release_tbf(tbf->control_ts->ulc, tbf);
+		tbf->control_ts = NULL;
+	}
 
 	/* Now simply detach from all attached PDCHs */
 	for (ts = 0; ts < 8; ts++) {
@@ -228,6 +235,12 @@ void tbf_unlink_pdch(struct gprs_rlcmac_tbf *tbf)
 		tbf->pdch[ts]->detach_tbf(tbf);
 		tbf->pdch[ts] = NULL;
 	}
+
+	/* Detach from TRX: */
+	if (tbf->trx) {
+		llist_del(tbf_trx_list(tbf));
+		tbf->trx = NULL;
+	}
 }
 
 void tbf_free(struct gprs_rlcmac_tbf *tbf)
@@ -259,7 +272,6 @@ void tbf_free(struct gprs_rlcmac_tbf *tbf)
 	tbf->stop_timers("freeing TBF");
 	/* TODO: Could/Should generate  bssgp_tx_llc_discarded */
 	tbf_unlink_pdch(tbf);
-	llist_del(tbf_trx_list(tbf));
 
 	if (tbf->ms())
 		tbf->set_ms(NULL);
@@ -295,10 +307,20 @@ void tbf_assign_control_ts(struct gprs_rlcmac_tbf *tbf)
 
 void gprs_rlcmac_tbf::n_reset(enum tbf_counters n)
 {
-	if (n >= N_MAX) {
-		LOGPTBF(this, LOGL_ERROR, "attempting to reset unknown counter %s\n",
-			get_value_string(tbf_counters_names, n));
-		return;
+	OSMO_ASSERT(n < N_MAX);
+
+	switch(n) {
+	case N3101:
+		OSMO_ASSERT(direction == GPRS_RLCMAC_UL_TBF);
+		break;
+	case N3103:
+		OSMO_ASSERT(direction == GPRS_RLCMAC_UL_TBF);
+		break;
+	case N3105:
+		OSMO_ASSERT(direction == GPRS_RLCMAC_DL_TBF);
+		break;
+	default:
+		break;
 	}
 
 	Narr[n] = 0;
@@ -309,22 +331,21 @@ bool gprs_rlcmac_tbf::n_inc(enum tbf_counters n)
 {
 	uint8_t chk;
 
-	if (n >= N_MAX) {
-		LOGPTBF(this, LOGL_ERROR, "attempting to increment unknown counter %s\n",
-			get_value_string(tbf_counters_names, n));
-		return true;
-	}
+	OSMO_ASSERT(n < N_MAX);
 
 	Narr[n]++;
 
 	switch(n) {
 	case N3101:
+		OSMO_ASSERT(direction == GPRS_RLCMAC_UL_TBF);
 		chk = bts->n3101;
 		break;
 	case N3103:
+		OSMO_ASSERT(direction == GPRS_RLCMAC_UL_TBF);
 		chk = bts->n3103;
 		break;
 	case N3105:
+		OSMO_ASSERT(direction == GPRS_RLCMAC_DL_TBF);
 		chk = bts->n3105;
 		break;
 	default:
@@ -437,8 +458,8 @@ void gprs_rlcmac_tbf::t_start(enum tbf_timers t, int T, const char *reason, bool
 		OSMO_ASSERT(false);
 	}
 
-	LOGPSRC(DTBF, LOGL_DEBUG, file, line, "%s %sstarting timer %s [%s] with %u sec. %u microsec, cur_fn=%d\n",
-	     tbf_name(this), osmo_timer_pending(&Tarr[t]) ? "re" : "",
+	LOGPSRC(DTBF, LOGL_DEBUG, file, line, "%s %starting timer %s [%s] with %u sec. %u microsec, cur_fn=%d\n",
+	     tbf_name(this), osmo_timer_pending(&Tarr[t]) ? "Res" : "S",
 	     get_value_string(tbf_timers_names, t), reason, sec, microsec, current_fn);
 
 	Tarr[t].data = this;
@@ -496,7 +517,7 @@ void gprs_rlcmac_tbf::poll_timeout(struct gprs_rlcmac_pdch *pdch, uint32_t poll_
 		}
 		bts_do_rate_ctr_inc(bts, CTR_RLC_ASS_TIMEDOUT);
 		bts_do_rate_ctr_inc(bts, CTR_PUA_POLL_TIMEDOUT);
-		if (n_inc(N3105)) {
+		if (this->direction == GPRS_RLCMAC_DL_TBF && n_inc(N3105)) {
 			osmo_fsm_inst_dispatch(this->state_fi, TBF_EV_MAX_N3105, NULL);
 			bts_do_rate_ctr_inc(bts, CTR_RLC_ASS_FAILED);
 			bts_do_rate_ctr_inc(bts, CTR_PUA_POLL_FAILED);
@@ -515,7 +536,7 @@ void gprs_rlcmac_tbf::poll_timeout(struct gprs_rlcmac_pdch *pdch, uint32_t poll_
 		}
 		bts_do_rate_ctr_inc(bts, CTR_RLC_ASS_TIMEDOUT);
 		bts_do_rate_ctr_inc(bts, CTR_PDA_POLL_TIMEDOUT);
-		if (n_inc(N3105)) {
+		if (this->direction == GPRS_RLCMAC_DL_TBF && n_inc(N3105)) {
 			osmo_fsm_inst_dispatch(this->state_fi, TBF_EV_MAX_N3105, NULL);
 			bts_do_rate_ctr_inc(bts, CTR_RLC_ASS_FAILED);
 			bts_do_rate_ctr_inc(bts, CTR_PDA_POLL_FAILED);
@@ -539,11 +560,11 @@ void gprs_rlcmac_tbf::poll_timeout(struct gprs_rlcmac_pdch *pdch, uint32_t poll_
 		dl_tbf = tbf_as_dl_tbf(this);
 		/* POLL Timeout expecting DL ACK/NACK: implies direction == GPRS_RLCMAC_DL_TBF */
 		OSMO_ASSERT(dl_tbf);
-		if (!(dl_tbf->state_fsm.state_flags & (1 << GPRS_RLCMAC_FLAG_TO_DL_ACK))) {
+		if (!dl_tbf->m_last_dl_poll_ack_lost) {
 			LOGPTBF(this, LOGL_NOTICE,
 				"Timeout for polling PACKET DOWNLINK ACK: %s\n",
 				tbf_rlcmac_diag(dl_tbf));
-			dl_tbf->state_fsm.state_flags |= (1 << GPRS_RLCMAC_FLAG_TO_DL_ACK);
+			dl_tbf->m_last_dl_poll_ack_lost = true;
 		}
 		if (dl_tbf->state_is(TBF_ST_RELEASING))
 			bts_do_rate_ctr_inc(bts, CTR_RLC_REL_TIMEDOUT);
@@ -567,46 +588,6 @@ void gprs_rlcmac_tbf::poll_timeout(struct gprs_rlcmac_pdch *pdch, uint32_t poll_
 	}
 }
 
-int gprs_rlcmac_tbf::setup(int8_t use_trx, bool single_slot)
-{
-	int rc;
-
-	if (ms_mode(m_ms) != GPRS)
-		enable_egprs();
-
-	/* select algorithm */
-	rc = the_pcu->alloc_algorithm(bts, this, single_slot, use_trx);
-	/* if no resource */
-	if (rc < 0) {
-		LOGPTBF(this, LOGL_NOTICE,
-			"Timeslot Allocation failed: trx = %d, single_slot = %d\n",
-			use_trx, single_slot);
-		bts_do_rate_ctr_inc(bts, CTR_TBF_ALLOC_FAIL);
-		return -1;
-	}
-	/* assign initial control ts */
-	tbf_assign_control_ts(this);
-
-	/* set timestamp */
-	osmo_clock_gettime(CLOCK_MONOTONIC, &meas.rssi_tv);
-
-	LOGPTBF(this, LOGL_INFO,
-		"Allocated: trx = %d, ul_slots = %02x, dl_slots = %02x\n",
-		this->trx->trx_no, ul_slots(), dl_slots());
-
-	m_ctrs = rate_ctr_group_alloc(this, &tbf_ctrg_desc, next_tbf_ctr_group_id++);
-	if (!m_ctrs) {
-		LOGPTBF(this, LOGL_ERROR, "Couldn't allocate TBF counters\n");
-		return -1;
-	}
-
-	tbf_update_state_fsm_name(this);
-
-	ms_attach_tbf(m_ms, this);
-
-	return 0;
-}
-
 const char *tbf_name(const gprs_rlcmac_tbf *tbf)
 {
 	return tbf ? tbf->name() : "(no TBF)";
@@ -624,8 +605,7 @@ const char *gprs_rlcmac_tbf::name(bool enclousure) const
 		OSMO_STRBUF_PRINTF(sb, ":TFI-%u-%u-%d",
 				   this->trx->bts->nr, this->trx->trx_no, tfi);
 	}
-	OSMO_STRBUF_PRINTF(sb, ":STATE-%s:%sGPRS",
-			   state_name(), is_egprs_enabled() ? "E" : "");
+	OSMO_STRBUF_PRINTF(sb, ":%c", is_egprs_enabled() ? 'E' : 'G');
 	if (m_ms) {
 		uint32_t tlli = ms_tlli(m_ms);
 		if (ms_imsi_is_valid(m_ms))
@@ -634,7 +614,7 @@ const char *gprs_rlcmac_tbf::name(bool enclousure) const
 			OSMO_STRBUF_PRINTF(sb, ":TLLI-0x%08x", tlli);
 	}
 	if (enclousure)
-		OSMO_STRBUF_PRINTF(sb, ")");
+		OSMO_STRBUF_PRINTF(sb, "){%s}", state_name());
 	return m_name_buf;
 }
 
@@ -818,7 +798,7 @@ int tbf_check_polling(const struct gprs_rlcmac_tbf *tbf, const struct gprs_rlcma
 
 	if (!tbf_is_control_ts(tbf, pdch)) {
 		char buf[128];
-		LOGPTBF(tbf, LOGL_DEBUG, "Polling cannot be "
+		LOGPTBF(tbf, LOGL_NOTICE, "Polling cannot be "
 			"scheduled in this TS %s (control TS %s)\n",
 			pdch_name(pdch),
 			tbf->control_ts ? pdch_name_buf(tbf->control_ts, buf, sizeof(buf)) : "none");
@@ -905,7 +885,7 @@ const char* tbf_rlcmac_diag(const struct gprs_rlcmac_tbf *tbf)
 			OSMO_STRBUF_PRINTF(sb, "Assignment was on CCCH|");
 		if (dl_tbf->state_fsm.state_flags & (1 << GPRS_RLCMAC_FLAG_PACCH))
 			OSMO_STRBUF_PRINTF(sb, "Assignment was on PACCH|");
-		if (dl_tbf->state_fsm.state_flags & (1 << GPRS_RLCMAC_FLAG_DL_ACK))
+		if (dl_tbf->m_first_dl_ack_rcvd)
 			OSMO_STRBUF_PRINTF(sb, "Downlink ACK was received|");
 		else
 			OSMO_STRBUF_PRINTF(sb, "No downlink ACK received yet|");
