@@ -21,6 +21,7 @@
 #include <gprs_pcu.h>
 
 struct gprs_rlcmac_tbf;
+struct gprs_rlcmac_pdch;
 
 enum tbf_dl_ass_fsm_event {
 	TBF_DL_ASS_EV_SCHED_ASS, /* Tx Uplink Assignment is pending */
@@ -40,26 +41,18 @@ struct tbf_dl_ass_fsm_ctx {
 	struct gprs_rlcmac_tbf* tbf; /* back pointer */
 };
 
-extern const struct osmo_tdef_state_timeout tbf_dl_ass_fsm_timeouts[32];
-/* Transition to a state, using the T timer defined in tbf_dl_ass_fsm_timeouts.
- * The actual timeout value is in turn obtained from conn->T_defs.
- * Assumes local variable fi exists. */
-#define tbf_dl_ass_fsm_state_chg(fi, NEXT_STATE) \
-	osmo_tdef_fsm_inst_state_chg(fi, NEXT_STATE, \
-				     tbf_dl_ass_fsm_timeouts, \
-				     the_pcu->T_defs, \
-				     -1)
-
 extern struct osmo_fsm tbf_dl_ass_fsm;
 
 
 /* passed as data in TBF_DL_ASS_EV_CREATE_RLCMAC_MSG */
 struct tbf_dl_ass_ev_create_rlcmac_msg_ctx {
+	const struct gprs_rlcmac_pdch *pdch; /* TS where the created DL ctrl block is to be sent */
 	uint32_t fn; /* FN where the created DL ctrl block is to be sent */
-	uint8_t ts; /* TS where the created DL ctrl block is to be sent */
 	struct msgb *msg; /* to be filled by FSM during event processing */
 };
 
 
-struct msgb *tbf_dl_ass_create_rlcmac_msg(const struct gprs_rlcmac_tbf* tbf, uint32_t fn, uint8_t ts);
-bool tbf_dl_ass_rts(const struct gprs_rlcmac_tbf* tbf);
+struct msgb *tbf_dl_ass_create_rlcmac_msg(const struct gprs_rlcmac_tbf *tbf,
+					  const struct gprs_rlcmac_pdch *pdch,
+					  uint32_t fn);
+bool tbf_dl_ass_rts(const struct gprs_rlcmac_tbf *tbf, const struct gprs_rlcmac_pdch *pdch);
